/*
* Copyright 2008 Free Software Foundation, Inc.
*
* This software is distributed under multiple licenses; see the COPYING file in the main directory for licensing information for this specific distribuion.
*
* This use of this software may be subject to additional restrictions.
* See the LEGAL file in the main directory for details.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

*/



#ifndef TRXMANAGER_H
#define TRXMANAGER_H

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "Threads.h"
#include "Sockets.h"
#include "Interthread.h"
#include "GSMCommon.h"
#include "GSMTransfer.h"
#include <list>
#include <Timeval.h>


/* Forward refs into the GSM namespace. */
namespace GSM {

class L1Decoder;

};


class ARFCNManager;


/**
	The TransceiverManager processes the complete transcevier interface.
	There is one of these for each access point.
 */
class TransceiverManager {

	private:

	/// the ARFCN managers under this TRX
	std::vector<ARFCNManager*> mARFCNs;		

	/// set true when the first CLOCK packet is received
	volatile bool mHaveClock;
	/// socket for clock management messages
	UDPSocket mClockSocket;		
	/// a thread to monitor the global clock socket
	Thread mClockThread;	
	Mutex mControlLock;			///< lock to prevent overlapping transactions
	UDPSocket mControlSocket;		///< socket for radio control
	std::string mInitData;			///< Init data (for debug)
        
        bool mStopRequest;                      ///< true if the thread is requested to stop

	bool mExitRecv;                         ///< Exiting received from lower layer
	bool m_statistics;                      ///< Statistics are enabled (BTS started)
        
        int mScanTable[GSM::sGsmMaxArfcns];     ///< uplink interference scan results
        static const unsigned mScanStepMs = 100;


	public:

	enum ChannelCombination {
		FILL,               ///< Channel is transmitted, but unused
		I,                  ///< TCH/FS
		II,                 ///< TCH/HS, idle every other slot
		III,                ///< TCH/HS
		IV,                 ///< FCCH+SCH+CCCH+BCCH, uplink RACH
		V,                  ///< FCCH+SCH+CCCH+BCCH+SDCCH/4+SACCH/4, uplink RACH+SDCCH/4
		VI,                 ///< CCCH+BCCH, uplink RACH
		VII,                ///< SDCCH/8 + SACCH/8
		NONE,               ///< Channel is inactive, default
		LOOPBACK,           ///< similar go VII, used in loopback testing
		IGPRS               ///< GPRS channel, like I but static filler frames.
	};

	// This magic flag is ORed with the TN TimeSlot in vectors passed to the transceiver
	// to indicate the radio block is a filler frame instead of a radio frame.
	// Must be higher than any possible TN.
	enum TransceiverFlags {
		SET_FILLER_FRAME = 0x10
	};

	/**
		Construct a TransceiverManager.
		@param numARFCNs Number of ARFCNs supported by the transceiver.
		@param wTRXAddress IP address of the transceiver.
		@param wBasePort The base port for the interface, as defined in README.TRX.
		@param wLocalAddr Local address
	*/
	TransceiverManager(int numARFCNs, int wBasePort,
		const char* wTRXAddress = "127.0.0.1",
		const char* wLocalAddr = "127.0.0.1");

	/**@name Accessors. */
	//@{
	ARFCNManager* ARFCN(unsigned i) { assert(i<mARFCNs.size()); return mARFCNs.at(i); }
	//@}

	bool haveClock() const { return mHaveClock; }
	bool exiting() const { return mExitRecv; }

	unsigned C0() const;
	unsigned numARFCNs() const { return mARFCNs.size(); }

	/** Block until the clock is set over the UDP link. */
	//void waitForClockInit() const;

	/**
		Log init data
	*/
	void logInit();

	/** Start the clock management thread and all ARFCN managers. */
	void start();

	/**
		Send a command packet and get the response packet.
		@param cmdString The NULL-terminated command string to send.
                @param cmdName The command name, without parameters.
		@param response A buffer for the response packet, assumed to be char[MAX_PACKET_LENGTH].
		@return Length of the response or -1 on failure.
	*/
	int sendCommandPacket(const char* cmdString, const char* cmdName, char* response);

	bool sendCommand(const char* cmd, int* iParam = 0, const char* sParam = 0,
	    int* rspParam = 0, int arfcn = -1);

	/**
		Reset the transceiver
		@return true on success.
	*/
	inline bool reset()
	{
		int p = numARFCNs();
		return sendCommand("RESET",&p);
	}

	/**
		Stop the transceiver
		@return true on success.
	*/
	inline bool stop()
	{ mStopRequest = true; return sendCommand("STOP"); }

	/**
		Start/Stop statistics
		@param on Statistics state to set
		@return True on success.
	*/
	inline bool statistics(bool on)
	{
		m_statistics = on;
		return sendCommand("STATISTICS",0,on ? "ON" : "OFF");
	}

	/** Clock service loop. */
	friend void* ClockLoopAdapter(TransceiverManager*);
        
        void startScan();
        
        void stopScan();
        
        unsigned numScanUnchecked() const;
        
        unsigned scanStepMs() const
        { return mScanStepMs; }
        
        unsigned scanBestArfcn() const;
        
        const int* scanTable() const
        { return mScanTable; }
        
        float scanMeanRssi() const;
        
        void scanTableUpdate(unsigned arfcn, int rssi);
        

	private:

	/** Handler for messages on the clock interface. */
	void clockHandler();
        
        void scanTableClear();
};




void* ClockLoopAdapter(TransceiverManager *TRXm);




/**
	The ARFCN Manager processes transceiver functions for a single ARFCN.
	When we do frequency hopping, this will manage a full rate radio channel.
*/
class ARFCNManager {

	private:

	TransceiverManager &mTransceiver;

	Mutex mDataSocketLock;			///< lock to prevent contentional for the socket
	UDPSocket mDataSocket;			///< socket for data transfer
	Thread mRxThread;				///< thread to receive data from rx

	/**@name The demux table. */
	//@{
	Mutex mTableLock;
	static const unsigned maxModulus=51*26*4;	///< maximum unified repeat period
	GSM::L1Decoder* mDemuxTable[8][maxModulus];		///< the demultiplexing table for received bursts
	//@}

	unsigned mARFCN;						///< the current ARFCN
	unsigned int mArfcnPos;						///< ARFCN index
        
        unsigned mScanArfcn;                                            ///< current scan mode ARFCN
        bool mScanning;                                                 ///< true if scan mode active
        ::Timeval mTuneTime;                                            ///< time of most recent tune
        
	public:

	ARFCNManager(unsigned int wArfncPos, const char* wTRXAddress, int wBasePort,
		const char* wLocalAddr, TransceiverManager &wTRX);

	/** Start the uplink thread. */
	void start();

	unsigned ARFCN() const { return mARFCN; }

	 // (pat) This passes the message through to UDPSocket::write(),
	 // which maps to DatagramSocket::write() which does an immediate sendto() on the socket.
	 // (pat) Renamed overloaded function to clarify code.
	 // Culprit says who called us, for debugging.
	void writeHighSideTx(const GSM::TxBurst& burst,const char *culprit);


	/**@name Transceiver controls. */
	//@{

	/**
		Tune to a given ARFCN.
		@param wARFCN Target for tuning.
		@param c0 True if we are C0
		@return true on success.
	*/
	bool tune(int wARFCN, bool c0 = false);

	/**
		Tune to a given ARFCN, but with rx and tx on the same (downlink) frequency.
		@param wARFCN Target for tuning, using downlink frequeny.
		@return true on success.
	*/
	bool tuneLoopback(int wARFCN);

	/** Turn off the transceiver. */
	bool powerOff();

	/** Turn on the transceiver. */
	
	/**
		Turn on the transceiver.
		@param warn Warn if the transceiver fails to start
	*/
	bool powerOn(bool warn);

        /**     
		Set maximum expected delay spread.
		@param km Max network range in kilometers.
		@return true on success.
        */
        bool setMaxDelay(unsigned km);

        /**     
                Set radio receive gain.
                @param new desired gain in dB.
                @return new gain in dB.
        */
        signed setRxGain(signed dB);

        /**     
                Set radio transmit attenuation
                @param new desired attenuation in dB.
                @return new attenuation in dB.
        */
        signed setTxAtten(signed dB);

        /**     
                Set radio frequency offset
                @param new desired freq. offset
                @return new freq. offset
        */
        signed setFreqOffset(signed offset);


        /**
                Get noise level as RSSI.
                @return current noise level.
        */
        signed getNoiseLevel(void);

	/**
	Get factory calibration values.
	@return current eeprom values.
	*/
	signed getFactoryCalibration(const char * param);

	/**
		Set power wrt full scale.
		@param dB Power level wrt full power.
		@return true on success.
	*/
	bool setPower(int dB);

	/**
		Set TSC for all slots on the ARFCN.
		@param TSC TSC to use.
		@return true on success.
	*/
	bool setTSC(unsigned TSC);

 	/**
		Set the BSIC (including setting TSC for all slots on the ARFCN).
		@param BSIC BSIC to use.
		@return true on success.
	*/
	bool setBSIC(unsigned BSIC);

	/**
		Describe the channel combination on a given slot.
		@param TN The timeslot number 0..7.
		@param combo Channel combination, GSM 05.02 6.4.1, 0 for inactive.
		@return true on success.
	*/
	bool setSlot(unsigned TN, unsigned combo);

	/**
		Set the given slot to run the handover burst correlator.
		@param TN The timeslot number.
		@return true on succes.
	*/
	bool setHandover(unsigned TN);

	/**
		Clear the given slot to run the handover burst correlator.
		@param TN The timeslot number.
		@return true on success.
	*/
	bool clearHandover(unsigned TN);

        /**
                Run custom radio device command
                @param command Command to execute
		@return true on success.
	*/
        bool runCustom(const char* command);

	//@}


	/** Install a decoder on this ARFCN. */
	void installDecoder(GSM::L1Decoder* wL1);


        /// Start UL scan mode.
        void startScan();
        
        /// Stop UL scan mode.
        void stopScan();

	private:

	/** Action for reception. */
	void driveRx();

	/** Demultiplex and process a received burst. */
	void receiveBurst(const GSM::RxBurst&);

	/** Receiver loop. */
	friend void* ReceiveLoopAdapter(ARFCNManager*);

	/**
		Send a command with a parameter.
		@param command The command name.
		@param param The parameter for the command.
		@param responseParam Optional parameter returned
		@return The status code, 0 on success, -1 on local failure.
	*/
	int sendCommand(const char*command, const char*param, int *responseParam);

	/**
		Send a command with a parameter.
		@param command The command name.
		@param param The parameter for the command.
                @param responseParam Optional parameter returned
		@return The status code, 0 on success, -1 on local failure.
	*/
	int sendCommand(const char* command, int param, int *responseParam=NULL);

	/**
		Send a command with a string parameter.
		@param command The command name.
		@param param The string parameter(s).
		@return The status code, 0 on success, -1 on local failure.
	*/
	int sendCommand(const char* command, const char* param);


	/**
		Send a command with no parameter.
		@param command The command name.
		@return The status code, 0 on success, -1 on local failure.
	*/
	int sendCommand(const char* command);


};


/** C interface for ARFCNManager threads. */
void* ReceiveLoopAdapter(ARFCNManager*);


#endif
// vim: ts=4 sw=4
